﻿using System;
using System.Configuration;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Ipc;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using DLW.Properties;
using Microsoft.Win32;

namespace DLW
{
    class Program : MarshalByRefObject
    {
        //OSが日本語版か(デフォルト文字コードがSJIS(932)か否かで判断・・だけだと足りないので、UICultureでも判断)
        internal static readonly bool isJP = (Encoding.Default.CodePage == 932 && Thread.CurrentThread.CurrentUICulture.LCID == 1041 ? true : false);
        //バージョン番号なしのアプリケーション名
        internal static readonly string ProductName = (isJP ? Process.GetCurrentProcess().MainModule.FileVersionInfo.FileDescription : Application.ProductName);
        //バージョン番号つきのアプリケーション名
        internal static readonly string FileDescription = ProductName + " Ver " + Application.ProductVersion;
        //ipcのチャネル名(二重起動防止用、複数セッションに耐えられるようセッションIDも付加)
        private static readonly string IpcServerChannel = typeof(Program).Namespace + Application.ProductVersion
            + Process.GetCurrentProcess().SessionId;

        //アプリケーションアイコンをExeファイルから取り出しておく
        internal static Icon Icon; //Run直前に取り出す = Icon.ExtractAssociatedIcon(Application.ExecutablePath);

        internal static ContextMenuStrip cms; //右クリックメニュー
        internal static ToolStripMenuItem KeyLock; //キーロック操作用メニュー項目
        private static volatile Tray tray;
        private static volatile MainForm mainForm;

        /// <summary>
        /// アプリケーションのメイン エントリ ポイントです。
        /// </summary>
        [STAThread]
        static void Main()
        {
            Log("static void Main()");
            //未処理例外をログ記録するようイベントハンドラを設定
            Application.ThreadException += (sender, e) => { Log(e.Exception); };
            Thread.GetDomain().UnhandledException += (sender, e) => { Log(e.ExceptionObject as Exception); };
            //Mutex確保成否に応じ処理起動
            bool createdNew;
            Mutex mutex = new Mutex(true, FileDescription, out createdNew);
            if (createdNew)
            {   //成功：IPCサーバをSingleton起動
                IpcServerChannel channel = new IpcServerChannel(IpcServerChannel);
                ChannelServices.RegisterChannel(channel, true);
                RemotingConfiguration.RegisterWellKnownServiceType(typeof(Program), typeof(Program).Name, WellKnownObjectMode.Singleton);
            }
            else
            {   //失敗：すでにSingleton起動済なプロセスのメソッドを呼び出し、リフレッシュ＆バルーン通知を行って終了
                ChannelServices.RegisterChannel(new IpcClientChannel(), true);
                RemotingConfiguration.RegisterWellKnownClientType(typeof(Program), "ipc://" + IpcServerChannel + "/" + typeof(Program).Name);
                new Program().Show("設定変更／終了は、トレイアイコンの右クリックより行ってください。");
                return;
            }

            //===== 以下、起動処理本体

            //C#既定のスタイル設定
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            //バージョン更新チェック
            Log("void Main() - versionCheck");
            string startupText = null;
            try
            {
                string version = Settings.Default.Version;
                if (string.IsNullOrEmpty(version))
                {   //以前のバージョンの設定ファイルがあれば、取り込みを行う
                    try
                    {
                        version = Settings.Default.GetPreviousVersion("Version") as string;
                        if (string.IsNullOrEmpty(version) == false)
                        {
                            Settings.Default.Upgrade();
                            Settings.Default.Version = Application.ProductVersion;
                            startupText = "旧バージョン（" + version + "）の設定を引き継ぎました。";
                        }
                    }
                    catch (ConfigurationException) { } //引継ぎ失敗時はエラーを黙殺
                    if (string.IsNullOrEmpty(startupText))
                    {   //初回起動時表示/アップグレード失敗時表示
                        startupText = "ディスクドライブおよびLAN／無線LANのアクセスを" + Environment.NewLine + "監視し、タスクトレイに表示します。";
                        //バージョン番号の設定・・・はやめておく。(設定ウィンドウで保存操作を行うまで起動時にバルーン表示する仕様、ということにする)
                        //Settings.Default.Version = Application.ProductVersion;
                    }
                    Settings.Default.Save();
                }
            }
            catch (ConfigurationException ex)
            {   //バージョン設定取得エラー発生時、例外発生の原因となったファイル名を特定する
                while (ex != null && ex.Filename == null)
                {   //ファイル名がnullならさらにInnerExceptionを探査
                    ex = (ex.InnerException as ConfigurationException);
                }
                //エラーメッセージ出力
                MessageBox.Show("設定ファイルが破損しています。" + Environment.NewLine
                    + Environment.NewLine
                    + (ex != null ? ex.Filename : "(ファイル不明)") + Environment.NewLine
                    + Environment.NewLine
                    + "上記のファイルを修正ないし削除のうえ、再起動してください。" + Environment.NewLine
                    , ProductName + " 起動エラー", MessageBoxButtons.OK, MessageBoxIcon.Error);
                ////いずれにせよ、自プロセスは強制終了
                return;
            }
            //Readme存在チェック
            string readme = Path.Combine(Path.GetDirectoryName(Assembly.GetEntryAssembly().Location), "DLW_Readme.txt");
            //NotifyIcon用のメニューを作成
            Log("void Main() - cms");
            cms = new ContextMenuStrip();
            cms.Items.AddRange(new ToolStripItem[]
            {
                new ToolStripMenuItem((isJP ? "Webサイトを表示" : "Author's WebSite") + "(&W)", null, 
                    (sender, e) => { Process.Start("http://hp.vector.co.jp/authors/VA055804/DLw/"); })
                { 
                    ToolTipText = "ブラウザを起動し、DLWアクセスランプのサポートページを表示します。" 
                },
                new ToolStripMenuItem((isJP ? "Readmeを表示" : "Show Readme") + "(&R)", null, 
                    (sender, e) => { Process.Start(readme); })
                {
                    ToolTipText = "DLWアクセスランプの使い方・注意事項などの説明を表示します。",
                    Enabled = File.Exists(readme)
                },
                new ToolStripSeparator() //-----------------------------
                {
                    Enabled = false
                }, 
                KeyLock = new ToolStripMenuItem((isJP ? "キーロック" : "KeyLock ") + "ON/Off(&K)")
                {
                    Visible = false
                },
                new ToolStripMenuItem((isJP ? "累計表示" : "Statistics") + "(&S)", null, 
                    (sender, e) => { tray.ShowPopup(sender,new MouseEventArgs(MouseButtons.Left, 1, 0, 0, 0)); })
                {
                    ToolTipText = "監視開始時刻および経過時間、各ドライブ等の読込／書込バイト数を表示します。"
                        + Environment.NewLine + "（アイコン左クリックでも表示できます）"
                },
                new ToolStripMenuItem((isJP ? "リフレッシュ" : "Refresh") + "(&U)", null, Refresh)
                {
                    ToolTipText = "表示対象ドライブ／LAN／無線LANを全更新し、読込／書込バイト数をリセットします。"
                },
                new ToolStripSeparator() //-----------------------------
                {
                    Enabled = false
                }, 
                new ToolStripMenuItem((isJP ? "設定" : "Customize") + "(&T)", null, Show)
                {
                    ToolTipText = "設定カスタマイズ画面を表示します。"
                },
                new ToolStripMenuItem((isJP ? "終了" : "Exit") + "(&X)", null, 
                    (sender,e) => { Application.Exit();})
                {
                    ToolTipText = "プログラム常駐・アクセス監視を終了します。"
                }
            });
            //--------
            cms.PreviewKeyDown += (sender, e) =>
            {   // Alt+F4対策#1：Altキーが押されたらフォーカスをはずす
                if (e.Alt)
                {
                    SetForegroundWindow(GetDesktopWindow());
                }
            };
            cms.Closing += (sender, e) =>
            {   // Alt+F4対策#2：メニューを閉じた段階でトレイのNativeWindowがアクティブならフォーカスをはずす
                IntPtr handle = GetForegroundWindow();
                foreach (NativeWindow window in tray.NativeWindow)
                {
                    if (handle == window.Handle)
                    {
                        SetForegroundWindow(GetDesktopWindow());
                        return;
                    }
                }
            };
            //アプリケーション起動（ただしフォームは表示しない）------------
            //NotifyIconをここで作成（cmsのハンドルも確保しておく）
            IntPtr dummy = Program.cms.Handle;
            Refresh(startupText, null);
            //アイコン取り出し：ICon.ExtractAssociatedIconではネットワーク共有フォルダexeからアイコン取得できないため、Win32APIで取得する
            const uint SHGFI_ICON = 0x100;
            const uint SHGFI_LARGEICON = 0x0; // 'Large icon  
            SHFILEINFO shinfo = new SHFILEINFO();
            IntPtr hImg = SHGetFileInfo(Assembly.GetEntryAssembly().Location, 0, ref shinfo, (uint)Marshal.SizeOf(shinfo), SHGFI_ICON | SHGFI_LARGEICON);
            Program.Icon = Icon.FromHandle(shinfo.hIcon);
            //メッセージループ起動
            Log("void Main() - run");
            Application.ApplicationExit += ApplicationExit;
            SystemEvents.SessionEnding += ApplicationExit;
            Application.Run();
            Log("void Main() - run end");
            //mutex開放
            mutex.ReleaseMutex();
        }

        /// <summary>フォーカス制御要否判定用</summary>
        [DllImport("user32")]
        static internal extern IntPtr GetForegroundWindow();
        /// <summary>フォーカス制御用</summary>
        [DllImport("user32")]
        static internal extern IntPtr GetDesktopWindow();
        /// <summary>フォーカス制御用</summary>
        [DllImport("user32")]
        [return: MarshalAs(UnmanagedType.Bool)]
        static internal extern bool SetForegroundWindow(IntPtr hWnd);
        /// <summary>アイコン画像取得用</summary>
        [DllImport("shell32.dll")]
        private static extern IntPtr SHGetFileInfo(string pszPath, uint dwFileAttributes, ref SHFILEINFO psfi, uint cbSizeFileInfo, uint uFlags);
        [StructLayout(LayoutKind.Sequential)]
        private struct SHFILEINFO
        {
            public IntPtr hIcon;
            public IntPtr iIcon;
            public uint dwAttributes;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
            public string szDisplayName;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 80)]
            public string szTypeName;
        };

        /// <summary>Mutex確保失敗(すでに起動済)用のインスタンスメソッド：先に起動しているプロセスへ処理が投げられる</summary>
        public void Show(string message)
        {   //初期化が終了していれば、バルーンメッセージ表示
            if (tray != null)
            {    //トレイ表示更新を行う
                Refresh(message);
            }
        }
        /// <summary>トレイ表示更新</summary>
        internal static void Refresh(string message)
        {
            cms.BeginInvoke((MethodInvoker)(() =>
            {
                Refresh(message, null);
            }));
        }
        private static void Refresh(object sender, EventArgs e)
        {
            lock (cms)
            {
                if (tray != null)
                {
                    tray.Dispose();
                }
                tray = new Tray(sender as string);
            }
        }
        /// <summary>設定画面表示を行う</summary>
        internal static void Show(object sender, EventArgs e)
        {
            if (mainForm == null || mainForm.IsDisposed)
            {
                mainForm = new MainForm();
                mainForm.Show();
            }
            mainForm.WindowState = FormWindowState.Normal;
            mainForm.Activate();
        }

        /// <summary>リソース破棄</summary>
        private static void ApplicationExit(object sender, EventArgs e)
        {
            if (tray != null) { tray.Dispose(); }
            if (cms != null) { cms.Dispose(); }
            if (mainForm != null && mainForm.IsDisposed == false) { mainForm.Dispose(); }
            if (Icon != null) { Tray.DestroyIcon(Icon.Handle); Icon.Dispose(); }
        }


        /// <summary>
        /// 例外をエラーログ出力／メッセージ表示する
        /// </summary>
        /// <param name="ex"></param>
        internal static void Log(Exception ex)
        {
            //発生箇所特定
            string stacktrace = ex.StackTrace;
            string source1 = "不明";
            string source2 = "";
            try
            {   //
                int startPos = stacktrace.IndexOf(typeof(Program).Namespace + ".");
                int endPos = stacktrace.IndexOf(Environment.NewLine, startPos);
                if (endPos == -1) { endPos = stacktrace.Length; }
                source1 = stacktrace.Substring(startPos, endPos - startPos);
                endPos = stacktrace.LastIndexOf(Environment.NewLine, endPos);
                startPos = stacktrace.LastIndexOf(Environment.NewLine, endPos);
                if (startPos == -1) { startPos = 0; }
                source2 = " →" + stacktrace.Substring(startPos, endPos - startPos).Trim('\r', '\n', ' ');
            }
            catch (Exception) { }
            //ログ出力
            if (ex != null)
            {
                Log("error", ex.GetType().FullName + "\t" + ex.Message + "\t" + ex.Source + Environment.NewLine + ex.StackTrace + Environment.NewLine);
            }
            //ダイアログも表示する
            if (MessageBox.Show("内部エラーが発生しました。\n（ご迷惑をおかけします。よろしければバグ報告をお寄せください）"
                + "\n\nエラー内容：" + (ex == null ? "（不明）" : ex.Message)
                + "\nエラー箇所：" + source1 + source2
                + "\n\n" + ProductName + "を続行しますか？", ProductName + " エラー", MessageBoxButtons.YesNo, MessageBoxIcon.Error) == DialogResult.No)
            {
                ApplicationExit(null, null); //リソースの開放を行う
                Environment.Exit(1);
            }
        }
        /// <summary>
        /// 各種ログ出力
        /// </summary>
        /// <param name="type"></param>
        /// <param name="message"></param>
        internal static void Log(string type, string message)
        {
            string text = DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") + "\t[" + FileDescription + "]\t" + type + "\t" + message;
            try
            {   //UTF8でエラーログファイルを書き出す
                using (StreamWriter writer = new StreamWriter(GetDirectoryName() + @"\" + type + "log.txt", true))
                {
                    writer.WriteLine(text);
                    writer.Close();
                }
            }
            catch (IOException) { }
        }

        /// <summary>
        /// デバッグログをコンソール表示
        /// </summary>
        /// <param name="message"></param>
        [Conditional("DEBUG")]
        internal static void Log(string message)
        {
            Console.WriteLine(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") + " - " + message);
        }

        /// <summary>
        /// アプリケーションのデータディレクトリ把握。
        /// 末尾「\」つきのディレクトリ名を返す。
        /// </summary>
        static string GetDirectoryName()
        {
            string logFilePath = System.IO.Path.GetTempPath(); //失敗時はtempディレクトリで代用
            try
            {   //DLWの設定ファイルuser.configが格納されるディレクトリの根元を、System.configurationに頼らずに自前で取得
                string path = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData)
                    + @"\" + System.Reflection.Assembly.GetEntryAssembly().GetName().Name + @"\";
                if (Directory.Exists(path) == false)
                {
                    Directory.CreateDirectory(path);
                }
                logFilePath = path;
            }
            catch (Exception) { }
            return logFilePath;
        }
    }
}
