﻿using System;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;

namespace DLW
{
    /// <summary>
    /// ショートカットファイルの読み込み（実パス変換）・作成・削除処理を提供するユーティリティクラス
    /// </summary>
    class ShortcutFile : IDisposable
    {
        /// <summary>
        /// ショートカットファイル名を組み立てて返す。
        /// すでに存在する場合は存在ファイル名、存在しなければ該当ディレクトリ直下に作成した場合のファイル名を返す
        /// </summary>
        /// <param name="dirname"></param>
        /// <param name="shortcutName"></param>
        /// <param name="subdir"></param>
        /// <returns></returns>
        internal static string ToShortcutPath(string dirname, string shortcutName, bool subdir)
        {
            //サブディレクトリのショートカットをOKとするか
            SearchOption option = (subdir == true ? SearchOption.AllDirectories : SearchOption.TopDirectoryOnly);
            //ショートカットファイルを探索
            string[] fi = Directory.GetFiles(dirname, shortcutName + ".lnk", option);
            return (fi.Length > 0 ? fi[0] : Path.Combine(dirname, shortcutName + ".lnk"));
        }

        /// <summary>
        /// 引数指定のファイル名がショートカットファイルのようであればtrueを返す
        /// </summary>
        /// <param name="fullPath"></param>
        /// <returns></returns>
        internal static bool IsShortcutFile(string fullPath)
        {
            return fullPath.Substring(fullPath.LastIndexOf('.') + 1).Equals("lnk", StringComparison.OrdinalIgnoreCase);
        }

        /// <summary>
        /// 引数指定のショートカットファイルのパスを実パスに変換する（変換失敗時はブランクを返す）
        /// </summary>
        /// <param name="fullPath"></param>
        /// <returns></returns>
        internal static string ToRealPath(string fullPath)
        {
            try
            {
                using (ShortcutFile helper = new ShortcutFile(fullPath))
                {
                    return helper.ToRealPath();
                }
            }
            catch (Exception) { }
            return "";
        }
        /// <summary>
        /// 引数指定のショートカットファイルのパスを実パスに変換し、起動パラメータも返す（変換失敗時はブランクを返す）
        /// </summary>
        /// <param name="fullPath"></param>
        /// <returns></returns>
        internal static string ToRealPath(string fullPath, out string startArgs)
        {
            try
            {
                using (ShortcutFile helper = new ShortcutFile(fullPath))
                {
                    startArgs = helper.GetArguments();
                    return helper.ToRealPath();
                }
            }
            catch (Exception) { }
            startArgs = "";
            return "";
        }

        /// <summary>
        /// ショートカットファイルを作成する(OKならNull、エラー時はエラーメッセージを返す)
        /// </summary>
        /// <param name="dirname"></param>
        /// <param name="shortcutName"></param>
        /// <param name="execPath"></param>
        /// <param name="iconShortcutNo"></param>
        /// <param name="description"></param>
        /// <returns></returns>
        internal static string CreateShortcutFile(string dirname, string shortcutName, string execPath, int iconShortcutNo, string description)
        {
            try
            {
                using (ShortcutFile helper = new ShortcutFile(Path.Combine(dirname, shortcutName + ".lnk")))
                {
                    helper.CreateShortcutFile(execPath, iconShortcutNo, description);
                }
                return null;
            }
            catch (Exception ex)
            {
                return "ショートカットを作成できません：" + ex.Message;
            }
        }
        /// <summary>
        /// ショートカットファイルを削除する(OKならNull、エラー時はエラーメッセージを返す)
        /// </summary>
        /// <returns></returns>
        internal static string Delete(string fullPath)
        {
            try
            {
                FileInfo file = new FileInfo(fullPath);
                if (file.Exists)
                {
                    if ((file.Attributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly)
                    {
                        file.Attributes = FileAttributes.Normal;
                    }
                    file.Delete();
                    return null;
                }
                else
                {
                    return "ショートカットファイルは既に削除されています";
                }
            }
            catch (Exception ex)
            {
                return "ショートカットを削除できません：" + ex.Message;
            }
        }

        #region インスタンスメソッド部------------------------------
        //メンバ変数
        object shortcut;
        object wshShell;
        //コンストラクタ：メンバ変数の初期化を行う
        private ShortcutFile(string fullPath)
        {
            //ディレクトリ不存在なら例外
            if (Directory.Exists(Path.GetDirectoryName(fullPath)) == false)
            {
                throw new DirectoryNotFoundException(Path.GetDirectoryName(fullPath));
            }
            //オブジェクト生成（生成失敗しても無視する）
            try
            {
                Type wshShellType = Type.GetTypeFromProgID("WScript.Shell");
                wshShell = Activator.CreateInstance(wshShellType);
                shortcut = wshShellType.InvokeMember("CreateShortcut", BindingFlags.InvokeMethod, null, wshShell, new object[] { fullPath });
            }
            catch (Exception) { }
        }
        //終了処理       
        public void Dispose()
        {
            if (shortcut != null)
            {
                try { Marshal.FinalReleaseComObject(shortcut); }
                catch (Exception) { }
            }
            if (wshShell != null)
            {
                try { Marshal.FinalReleaseComObject(wshShell); }
                catch (Exception) { }
            }
        }
        //リンク先パス取得
        private string ToRealPath()
        {
            return shortcut.GetType().InvokeMember("TargetPath", BindingFlags.GetProperty, null, shortcut, null).ToString();
        }
        //起動時パラメータ取得
        private string GetArguments()
        {
            return shortcut.GetType().InvokeMember("Arguments", BindingFlags.GetProperty, null, shortcut, null).ToString();
        }
        //情報保存
        /// <summary>
        /// ショートカットを作成・保存する
        /// </summary>
        /// <returns>成功時true,失敗時false</returns>
        private void CreateShortcutFile(string execPath, int iconShortcutNo, string description)
        {
            shortcut.GetType().InvokeMember("TargetPath", BindingFlags.SetProperty, null, shortcut, new object[] { execPath });
            shortcut.GetType().InvokeMember("WorkingDirectory", BindingFlags.SetProperty, null, shortcut, new object[] { Environment.CurrentDirectory });
            shortcut.GetType().InvokeMember("WindowStyle", BindingFlags.SetProperty, null, shortcut, new object[] { 1 });
            shortcut.GetType().InvokeMember("Description", BindingFlags.SetProperty, null, shortcut, new object[] { description });
            shortcut.GetType().InvokeMember("IconLocation", BindingFlags.SetProperty, null, shortcut, new object[] { execPath + "," + iconShortcutNo.ToString("0") });
            shortcut.GetType().InvokeMember("Save", BindingFlags.InvokeMethod, null, shortcut, null);
        }
        #endregion
    }
}
