﻿using DLW.Properties;
using System;
using System.Diagnostics;
using System.IO;

namespace DLW
{
    partial class Watcher
    {
        /// <summary>
        /// ドライブ監視クラス（監視対象なしのものもこれで準用）
        /// </summary>
        private class Drive : Watcher
        {
            volatile Exception ex = null; //エラー有無表示用(ドライブ検出変更時にクリア)
            //ドライブ監視
            readonly new PerformanceCounter rd = null;
            readonly new PerformanceCounter wr = null;

            /// <summary>コンストラクタ：ドライブ監視オブジェクトも格納</summary>
            internal Drive(char drive)
                : base(drive)
            {
                if (drive == WUnknown)
                {
                    this.ex = new Exception(Program.isJP ? "監視対象なし" : "No Target");
                }
                else
                {
                    try
                    {
                        rd = new PerformanceCounter("LogicalDisk", "Disk Read Bytes/sec", drive + ":");
                        wr = new PerformanceCounter("LogicalDisk", "Disk Write Bytes/sec", drive + ":");
                    }
                    catch (Exception ex)
                    {
                        this.ex = ex;
                    }
                }
                Check(); //準備ができているかチェックする
            }
            //破棄
            public override void Dispose()
            {
                if (rd != null) { rd.Dispose(); }
                if (wr != null) { wr.Dispose(); }
            }

            /// <summary>サマリー表示テキストを返す</summary>
            internal override string Text
            {
                get
                {
                    if (ex == null)
                    {
                        return base.Text;
                    }
                    //監視不能な場合はその旨を返す
                    string t = (Program.isJP ? "監視不能(パフォーマンスカウンタ" : "PerformanceCounter Invalid(");
                    return Name + '\t' +
                        (floppyDrive.IndexOf(drive) > -1 ? "Floppy"
                        : ex is DriveNotFoundException ? "Not Ready"
                        : ex is InvalidOperationException ? t + "取得不能)"
                        : ex is FormatException ? t + "破損)"
                        : ex is UnauthorizedAccessException ? t + "参照権限なし)"
                        : "エラー(" + ex.Message + ")");
                }
            }

            /// <summary>ドライブの準備ができているかチェック（ドライブの例外を設定／クリア）</summary>
            internal override void Check()
            {
                //パフォーマンスカウンタが準備できている（有効なドライブ）の場合、例外設定を初期化
                if (rd != null && wr != null)
                {   //ドライブの準備ができていれば例外をnullに、できていなければその旨を設定
                    this.ex = (readyDrive.IndexOf(drive) == -1 ? new DriveNotFoundException() : null);
                }
            }
            /// <summary>read/write有無チェック:監視結果に応じた値を返す</summary>
            internal override int Check(int tick)
            {
                if (this.ex == null)
                {
                    try
                    {   //読込バイト数チェック
                        float tempRd = rd.NextValue();
                        if (tempRd > 0)
                        {
                            base.rd = tick;
                            rdCount += ((long)tempRd * Settings.Default.TTick + 500) / 1000;
                        };
                        //書込みバイト数チェック
                        float tempWr = wr.NextValue();
                        if (tempWr > 0)
                        {
                            base.wr = tick;
                            wrCount += ((long)tempWr * Settings.Default.TTick + 500) / 1000;
                        };
                        return base.Check(tick);
                    }
                    catch (Exception ex)
                    {   //このドライブは有効ではなくなった
                        this.ex = ex;
                    }
                }
                return 0; //無効状態
            }
        }
    }
}
