﻿using System;
using System.Collections.Generic;
using System.Net.NetworkInformation;
using DLW.Properties;

namespace DLW
{
    /// <summary>
    /// ネットワーク監視クラス
    /// </summary>
    class Network : Watcher
    {
        #region スタティックメンバ／メソッド------------------------------------
        /// <summary>NICの変更を検知して更新する（イベントなどからの汎用呼び出し、Trayのサスペンドからの復帰時にも呼び出し）</summary>
        internal static void Refresh(object sender, EventArgs e)
        {
            allNi = NetworkInterface.GetAllNetworkInterfaces();
        }
        //有効ネットワークインターフェースの一覧（変更検出イベントで更新）
        static volatile NetworkInterface[] allNi;
        static Network()
        {   //NICの変更（USBなどのNICの着脱）を検知して更新(両イベント仕込んでおく)
            Refresh(null, null);
            NetworkChange.NetworkAvailabilityChanged += Refresh;
            NetworkChange.NetworkAddressChanged += Refresh;
        }

        /// <summary>static:drive名→対象NICの一覧</summary>
        private static NetworkInterfaceType[] Get(char drive)
        {
            return
                (drive == WWireless ? new NetworkInterfaceType[]
                    {
                        NetworkInterfaceType.Wireless80211
                    }
                : drive == WEthernet ? new NetworkInterfaceType[]
                    {
                        NetworkInterfaceType.Ethernet,
                        NetworkInterfaceType.Ethernet3Megabit,
                        NetworkInterfaceType.FastEthernetFx,
                        NetworkInterfaceType.FastEthernetT,
                        NetworkInterfaceType.GigabitEthernet
                    }
                : new NetworkInterfaceType[0]);
        }
        /// <summary>static:指定されたネットワークインターフェースの有無を返す</summary>
        internal static bool Check(char drive)
        {
            NetworkInterfaceType[] nit = Get(drive);
            foreach (NetworkInterface n in allNi)
            {   //有効なインターフェースが見つかったらtrue
                if (Array.IndexOf(nit, n.NetworkInterfaceType) >= 0)
                {
                    return true;
                }
            }
            return false;
        }
        #endregion

        #region インスタンスメンバ／メソッド------------------------------------
        volatile NetworkInterface[] wholeList = null; //このオブジェクトが把握している全NIC（最新とズレがないか逐次チェックすること）
        NetworkInterface[] watchList = null; //このオブジェクトが監視対象とするNIC（NIC種類などより決定）
        IPv4InterfaceStatistics[] niStat = null; //監視対象NICの直前の統計情報
        bool IgnoreBroadcastRx; //ブロードキャスト受信を無視するか（監視オブジェクト生成時に設定）

        /// <summary>コンストラクタ</summary>
        internal Network(char drive)
            : base(drive)
        {
            IgnoreBroadcastRx = Settings.Default.IgnoreBroadcastRx;
        }

        /// <summary>表示名を返す</summary>
        internal override string Name
        {
            get
            {   //種類に応じた文字列を返す
                return drive == WEthernet ? "LAN"
                    : drive == WWireless ? (Program.isJP ? "無線" : "WLAN")
                    : drive == WSpecifiedLAN ? Settings.Default.SpecifiedLanName
                    : base.Name;
            }
        }

        /// <summary>サマリー表示テキストを返す</summary>
        internal override string Text
        {
            get
            {
                foreach (IPv4InterfaceStatistics stat in niStat)
                {   //ひとつでも有効な接続があれば通常表示内容を返す
                    if (stat != null) { return base.Text; }
                }
                //有効な接続なし：接続なしという文言をつける
                return base.Text + (Program.isJP ? " (現在接続なし)" : " (Not Connected)");
            }
        }

        /// <summary>read/write有無チェック:監視結果に応じた値を返す</summary>
        internal override int Check(int tick)
        {   //LAN更新チェック・監視対象確認
            if (object.ReferenceEquals(allNi, this.wholeList) == false)
            {   //参照が異なればオブジェクトが変わったものと判断し、監視対象を再確認する
                this.wholeList = allNi;
                List<NetworkInterface> watchList = new List<NetworkInterface>();
                NetworkInterfaceType[] nit = Get(this.drive);
                string keyword = (Contains(WSpecifiedLAN) ? Settings.Default.SpecifiedLanName : null);
                foreach (NetworkInterface ni in wholeList)
                {
                    if (string.IsNullOrEmpty(keyword) == false && (ni.Description.Contains(keyword) || ni.Name.Contains(keyword)))
                    {   //別表示LANのキーワードがNICのDescriptionに含まれる：このオブジェクトが別表示LANであれば対象に含める
                        if (this.drive == WSpecifiedLAN && ni.Supports(NetworkInterfaceComponent.IPv4)) { watchList.Add(ni); }
                    }
                    else
                    {   //そうでなければ、NIC種類がこのオブジェクトの監視対象に含まれていれば対象とする
                        if (Array.IndexOf(nit, ni.NetworkInterfaceType) >= 0) { watchList.Add(ni); }
                    }
                }
                this.watchList = watchList.ToArray(); //監視対象を配列化する。直後に統計情報配列も確保する
                this.niStat = new IPv4InterfaceStatistics[this.watchList.Length];
            }
            bool ret = false;
            IPv4InterfaceStatistics current;
            IPv4InterfaceStatistics prev;
            for (int i = 0; i < this.watchList.Length; i++)
            {
                NetworkInterface ni = this.watchList[i];
                if (ni.OperationalStatus == OperationalStatus.Up)
                {
                    current = ni.GetIPv4Statistics();
                    prev = niStat[i];
                    niStat[i] = current;
                    if (prev != null)
                    {
                        //受信バイト(orパケット)の変化があるかチェック
                        if (current.UnicastPacketsReceived - prev.UnicastPacketsReceived > 0
                            || (IgnoreBroadcastRx == false && current.BytesReceived - prev.BytesReceived > 0))
                        {
                            rdCount += current.BytesReceived - prev.BytesReceived;
                            rd = tick;
                        }
                        //送信バイトの変化があるかチェック
                        long bytesSent = current.BytesSent - prev.BytesSent;
                        if (bytesSent > 0)
                        {
                            wrCount += bytesSent;
                            wr = tick;
                        }
                    }
                    ret = true;
                }
                else
                {
                    niStat[i] = null;
                }
            }
            //接続ありのnitがあれば値を返す。そうでなければ無効
            return (ret ? base.Check(tick) : 0);
        }
        #endregion
    }
}