﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Management;
using System.Runtime.InteropServices;
using System.Text;
using DLW.Properties;

namespace DLW
{
    /// <summary>
    /// 各ドライブの監視オブジェクト基底クラス
    /// </summary>
    abstract partial class Watcher : IDisposable
    {
        #region 定数定義--------------------------------------------------------
        //ドライブ監視定義文字列
        internal const char WDisk = 'D';
        internal const char WRemovable = 'M';
        internal const char WRamDisk = 'R';
        //ネットワーク監視定義文字列
        internal const char WEthernet = 'l';
        internal const char WWireless = 'w';
        internal const char WSpecifiedLAN = 'v'; //一部のLANを別表示
        //キーロック監視/ロック定義文字列
        internal const char WKeylock = 'k';
        internal const char XNumLk = 'n'; //0,1,2=ロックOn/Off/無指定
        internal const char XScrLk = 's'; //0,1,2=ロックOn/Off/無指定
        internal const char XCapsLock = 'c'; //0,1,2=ロックOn/Off/無指定
        //その他
        internal const char WCustom = '!';
        internal const char WAggregateDisk = '_'; //パーティション集約表示。0=ドライブ名、1=１開始連番、2=0開始のディスク番号
        private const char WUnknown = '?';

        //キーロック指定等汎用
        /// <summary>targetに指定charが含まれる場合、その直後の数字設定（ロックOn/Off/無指定、など）を返す</summary>
        internal static int IndexOf(char drive)
        {
            int index = Settings.Default.Target.IndexOf(drive);
            if (index >= 0 && Settings.Default.Target.Length > index + 1)
            {
                char ch = Settings.Default.Target[index + 1];
                if (ch == '1') { return 1; }
                else if (ch == '2') { return 2; }
            }
            return 0;
        }        
        #endregion

        #region スタティックメンバ／メソッド------------------------------------
        //マウントされたドライブ／LANの一覧（パフォーマンスカウンタで読み出しできないドライブも含む。staticのRefresh呼び出し時に更新）
        static volatile string readyDrive = "";
        //フロッピードライブの一覧
        static volatile string floppyDrive = "";

        /// <summary>ドライブ/ネットワーク着脱更新チェック（把握しているドライブ／LAN一覧も更新）</summary>
        internal static bool Refresh()
        {
            StringBuilder sb = new StringBuilder();

            //ドライブの検出
            foreach (DriveInfo di in DriveInfo.GetDrives())
            {
                //有効ドライブかどうか調査
                char drive = di.Name[0];
                if (di.DriveType == DriveType.Removable && readyDrive.IndexOf(drive) == -1)
                {   //初出orフロッピードライブの特殊判定
                    if (floppyDrive.IndexOf(drive) == -1)
                    {   //まだフロッピーと判定されていない
                        try
                        {
                            ManagementObject mo = new ManagementObject("win32_logicaldisk.deviceid=" + '"' + drive + ':' + '"');
                            mo.Get();
                            uint? mediaType = (mo["MediaType"] as uint?); //HDDなら12,USBならNull。FDDならそれっぽい値になりそう。MOは不明・・・
                            if (mediaType != null && mediaType != 0 && mediaType != 11 && mediaType != 12)
                            {   //このドライブはフロッピーである。isReadyを調査するとまずいことになるので次のループへ
                                floppyDrive += drive.ToString();
                                continue;
                            }
                        }
                        catch (ManagementException)
                        {   //logicaldiskの情報が取れないドライブは有効ではないとみなす
                            continue;
                        }
                        catch (COMException ex)
                        {
                            if ((uint)ex.ErrorCode == 0x80010002)
                            {   //RPC_E_CALL_CANCELED：WMIから情報取得できない（Windowsログオン直後はまれによくあるらしい）
                                continue;
                            }
                            throw ex;
                        }
                    }
                    else
                    {   //すでにフロッピーと判定済。isReadyを調査するとまずいことになるので次のループへ
                        continue;
                    }
                }
                if (di.IsReady) { sb.Append(drive); } //ドライブ名の先頭１文字を格納
            }
            //ネットワークの検出
            if (Network.Check(WEthernet)) { sb.Append(WEthernet); }
            if (Network.Check(WWireless)) { sb.Append(WWireless); }
            //相違あればリフレッシュ
            if (readyDrive != sb.ToString())
            {
                readyDrive = sb.ToString();
                return true;
            }
            return false;
        }

        /// <summary>static:監視オブジェクトを一括生成して返す</summary>
        internal static Watcher[] Get(string drive)
        {
            //物理ディスク集約用ディクショナリ確保
            Dictionary<string, Watcher> disks = null;
            if (Watcher.Contains(WAggregateDisk))
            {
                try
                {
                    string[] names = new PerformanceCounterCategory("PhysicalDisk").GetInstanceNames();
                    disks = new Dictionary<string, Watcher>(names.Length);
                    foreach (string name in names) { disks.Add(name, null); }
                }
                catch (Exception) { }
            }
            //
            List<Watcher> watcher = new List<Watcher>();
            for (int i = 0; i < drive.Length; i++)
            {
                Program.Log("Watcher.get() - watcher生成:" + drive[i]);

                switch (drive[i])
                {
                    case WEthernet:
                    case WWireless:
                    case WSpecifiedLAN:
                        watcher.Add(new Network(drive[i]));
                        break;
                    case WKeylock:
                        watcher.Add(new KeyLock());
                        break;
                    default:
                        Watcher w = new Watcher.Drive(drive[i]);
                        Watcher already = null;
                        if (disks != null)
                        {   //物理ディスク集約要か調べる
                            foreach (string name in disks.Keys)
                            {
                                if (name.Contains(w.Name))
                                {
                                    already = disks[name];
                                    if (already == null)
                                    {   //初出：単純追加にあたり物理ディスクのディクショナリへ登録
                                        disks[name] = w;
                                    }
                                    else
                                    {   //物理ディスクのディクショナリに登録されている場合
                                        Disk d = (already as Disk);
                                        if (d == null)
                                        {   //未集約なので該当ドライブを物理ディスクに集約する
                                            int pattern = Watcher.IndexOf(WAggregateDisk);
                                            char driveletter = //パターン:0=先頭ドライブ名、1=1開始、2=0開始
                                                (pattern == 0 ? already.drive
                                                : (char)(name[0] + (pattern == 1 ? 1 : 0))); 
                                            d = new Disk(driveletter, already);
                                            disks[name] = d;
                                            int idx = watcher.IndexOf(already);
                                            watcher[idx] = d;
                                        }
                                        d.Add(w);
                                    }
                                    break; //foreach
                                }
                            }
                        }
                        if (already == null)
                        {   //物理ディスク集約対象外：追加する
                            watcher.Add(w);
                        }
                        break;
                }
            }
            return watcher.ToArray();
        }
        /// <summary>static:監視対象のドライブ/LANの一覧を取得</summary>
        /// <returns></returns>
        internal static String Get()
        {
            //ドライブの一覧・ネットワークの一覧を更新しておく
            Program.Log("Watcher.get(引数なし):開始計測");
            Refresh();
            Program.Log("Watcher.get(引数なし):Driveリフレッシュ後");
            Network.Refresh(null, null);
            Program.Log("Watcher.get(引数なし):Networkリフレッシュ後");
            //監視対象の一覧を把握,stringbuilder化
            StringBuilder sb = new StringBuilder();
            //キーロック
            if (Contains(WKeylock)) { sb.Append(WKeylock); }
            //ドライブ
            sb.Append(Get(Settings.Default.Target, Settings.Default.TargetOmit, Settings.Default.TargetCustom));
            //監視対象LANを把握
            if (Contains(WSpecifiedLAN)) { sb.Append(WSpecifiedLAN); } //特定LANを先頭にする
            if (Contains(WEthernet) && Network.Check(WEthernet)) { sb.Append(WEthernet); }
            if (Contains(WWireless) && Network.Check(WWireless)) { sb.Append(WWireless); }

            if (sb.Length == 0)
            {   //有効ドライブなし
                return WUnknown.ToString();
            }
            return sb.ToString();
        }
        /// <summary>static:監視対象のドライブ一覧を取得</summary>
        internal static String Get(String target, String omit, String custom)
        {
            string ret = "";
            if (target.IndexOf(WCustom) >= 0)
            {   //カスタム指定に基づき対象ドライブを把握
                foreach (char ch in custom)
                {
                    if (ret.IndexOf(ch) == -1 && ch >= 'A' && ch <= 'Z')
                    {
                        ret += ch;
                    }
                }
            }
            else
            {
                foreach (DriveInfo di in DriveInfo.GetDrives())
                {
                    char ch = di.Name[0]; //ドライブ名先頭1文字を取得
                    //有効ドライブかどうか調査(RemovableはFloppyでないことも条件に加える)
                    if (omit.IndexOf(ch) == -1 &&
                        (di.DriveType == DriveType.Fixed && target.IndexOf(WDisk) >= 0
                            || (di.DriveType == DriveType.Removable && target.IndexOf(WRemovable) >= 0 && floppyDrive.IndexOf(ch) == -1)
                            || (di.DriveType == DriveType.Ram && target.IndexOf(WRamDisk) >= 0)
                        ) && di.IsReady)
                    {
                        ret += ch;
                    }
                }
            }
            return ret;
        }
        /// <summary>static汎用:targetに指定ドライブが含まれるかを返す</summary>
        internal static bool Contains(char drive)
        {
            return (Settings.Default.Target.IndexOf(drive) >= 0);
        }

        #endregion

        #region インスタンスメンバ／メソッド------------------------------------
        readonly internal char drive;
        protected int rd; //アクセス有無把握・延長表示用
        protected int wr; //アクセス有無把握・延長表示用
        protected long rdCount = 0; //読み込みバイト数表示用
        protected long wrCount = 0; //書き込みバイト数表示用
        int TDelay; //延長表示時間（監視オブジェクト生成時に設定）

        /// <summary>コンストラクタ：ドライブ情報のみ格納</summary>
        protected Watcher(char drive)
        {
            this.drive = drive;
            TDelay = Settings.Default.TDelay;
            rd = Environment.TickCount - TDelay;
            wr = rd;
        }
        /// <summary>オブジェクト破棄</summary>
        public virtual void Dispose() { }

        /// <summary>表示名を返す</summary>
        internal virtual string Name
        {
            get { return drive + ":"; }
        }

        /// <summary>サマリー表示テキストを返す</summary>
        internal virtual string Text
        {
            get
            {
                return Name + '\t' + rdCount.ToString("#,##0") + " / " + wrCount.ToString("#,##0");
            }
        }


        /// <summary>エラー情報などがあればリフレッシュ</summary>
        internal virtual void Check() { }

        /// <summary>read/write有無チェック:監視結果に応じた値を返す</summary>
        internal virtual int Check(int tick)
        {
            //読み出し状況に応じたアイコン決定(読み出し中=3,直近読み出し=2,否なら1)
            //※tickは24.9日で符号反転する。読み書き実績のないドライブの外枠表示を抑止するよう対応
            int iconNo =
                (tick == rd) ? 3 :
                (tick - rd < TDelay && rdCount > 0) ? 2
                : 1;
            //書き込み状況に応じたアイコン決定
            if (tick == wr)
            {   //書き込みあり(+6)
                iconNo += 6;
            }
            else if (tick - wr < TDelay && wrCount > 0)
            {   //直近に書き込み実施(+3)
                iconNo += 3;
            }
            return iconNo;
        }
        #endregion
    }
}
